/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CPU_OPERATOR_MATMULIMPL_FORWARD_KERNEL_H_
#define AIDGE_CPU_OPERATOR_MATMULIMPL_FORWARD_KERNEL_H_

#include "aidge/utils/Registrar.hpp"
#include <algorithm>

#include "aidge/backend/cpu/operator/MatMulImpl.hpp"

namespace Aidge {

template <class I, class W, class O>
void MatMulImpl_cpu_forward_kernel(const MatMul_Op::Attrs& attrs, const DimSize_t batchSize, const DimSize_t oneInputSize,
                                   const void* input_, const void* weights_, void* output_) {
    // FIXME: missing MatMul parameters as arguments
    const I* input = static_cast<const I*>(input_);
    const W* weights = static_cast<const W*>(weights_);
    O* output = static_cast<O*>(output_);


    std::fill(output, output+(batchSize*std::get<0>(attrs)), O(0));

    for (std::size_t batch = 0; batch < batchSize; ++batch) {
        for (std::size_t out = 0; out < std::get<0>(attrs); ++out) {
            output[out + batch*std::get<0>(attrs)] = std::inner_product(input + batch*oneInputSize,
                                                        input + (batch + 1)*oneInputSize,
                                                        weights + out*oneInputSize,
                                                        output[out + batch*std::get<0>(attrs)]);
        }
    }
}


namespace {
static Registrar<MatMulImplForward_cpu> registrarMatMulImpl2DForward_cpu_Float32(
        {DataType::Float32, DataType::Float32, DataType::Float32},
        Aidge::MatMulImpl_cpu_forward_kernel<float, float, float>);
static Registrar<MatMulImplForward_cpu> registrarMatMulImpl2DForward_cpu_Int32(
        {DataType::Int32, DataType::Int32, DataType::Int32},
        Aidge::MatMulImpl_cpu_forward_kernel<int, int, int>);
static Registrar<MatMulImplForward_cpu> registrarMatMulImpl2DForward_cpu_Float64(
        {DataType::Float64, DataType::Float64, DataType::Float64},
        Aidge::MatMulImpl_cpu_forward_kernel<double, double, double>);
}  // namespace

}  // namespace Aidge

#endif /* AIDGE_CPU_OPERATOR_MATMULIMPL_FORWARD_KERNEL_H_ */
